@@ -17,7 +17,6 @@ from sqlalchemy import text
 from sqlalchemy import update
 from sqlalchemy.orm import backref
 from sqlalchemy.orm import joinedload
-from sqlalchemy.orm import mapper
 from sqlalchemy.orm import relationship
 from sqlalchemy.orm import Session
 from sqlalchemy.orm import sessionmaker
@@ -87,7 +86,7 @@ class UpdateDeleteTest(fixtures.MappedTest):
         Address = cls.classes.Address
         addresses = cls.tables.addresses
 
-        mapper(
+        cls.mapper_registry.map_imperatively(
             User,
             users,
             properties={
@@ -95,7 +94,40 @@ class UpdateDeleteTest(fixtures.MappedTest):
                 "addresses": relationship(Address),
             },
         )
-        mapper(Address, addresses)
+        cls.mapper_registry.map_imperatively(Address, addresses)
+
+    @testing.combinations("table", "mapper", "both", argnames="bind_type")
+    @testing.combinations(
+        "update", "insert", "delete", argnames="statement_type"
+    )
+    def test_get_bind_scenarios(self, connection, bind_type, statement_type):
+        """test for #7936"""
+
+        User = self.classes.User
+
+        if statement_type == "insert":
+            stmt = insert(User).values(
+                {User.id: 5, User.age: 25, User.name: "spongebob"}
+            )
+        elif statement_type == "update":
+            stmt = (
+                update(User)
+                .where(User.id == 2)
+                .values({User.name: "spongebob"})
+            )
+        elif statement_type == "delete":
+            stmt = delete(User)
+
+        binds = {}
+        if bind_type == "both":
+            binds = {User: connection, User.__table__: connection}
+        elif bind_type == "mapper":
+            binds = {User: connection}
+        elif bind_type == "table":
+            binds = {User.__table__: connection}
+
+        with Session(binds=binds) as sess:
+            sess.execute(stmt)
 
     def test_illegal_eval(self):
         User = self.classes.User
@@ -201,7 +233,7 @@ class UpdateDeleteTest(fixtures.MappedTest):
                 return User.name.__clause_element__()
 
         s = fixture_session()
-        jill = s.query(User).get(3)
+        jill = s.get(User, 3)
         s.query(User).update(
             {Thing(): "moonbeam"}, synchronize_session="evaluate"
         )
@@ -228,7 +260,7 @@ class UpdateDeleteTest(fixtures.MappedTest):
         User = self.classes.User
 
         s = fixture_session()
-        jill = s.query(User).get(3)
+        jill = s.get(User, 3)
         s.query(User).update(
             {column("name"): "moonbeam"}, synchronize_session="evaluate"
         )
@@ -240,10 +272,12 @@ class UpdateDeleteTest(fixtures.MappedTest):
         class Foo(object):
             pass
 
-        mapper(Foo, self.tables.users, properties={"uname": synonym("name")})
+        self.mapper_registry.map_imperatively(
+            Foo, self.tables.users, properties={"uname": synonym("name")}
+        )
 
         s = fixture_session()
-        jill = s.query(Foo).get(3)
+        jill = s.get(Foo, 3)
         s.query(Foo).update(
             {"uname": "moonbeam"}, synchronize_session="evaluate"
         )
@@ -253,10 +287,12 @@ class UpdateDeleteTest(fixtures.MappedTest):
         class Foo(object):
             pass
 
-        mapper(Foo, self.tables.users, properties={"uname": synonym("name")})
+        self.mapper_registry.map_imperatively(
+            Foo, self.tables.users, properties={"uname": synonym("name")}
+        )
 
         s = fixture_session()
-        jill = s.query(Foo).get(3)
+        jill = s.get(Foo, 3)
         s.query(Foo).update(
             {Foo.uname: "moonbeam"}, synchronize_session="evaluate"
         )
@@ -266,14 +302,14 @@ class UpdateDeleteTest(fixtures.MappedTest):
         class Foo(object):
             pass
 
-        mapper(
+        self.mapper_registry.map_imperatively(
             Foo,
             self.tables.users,
             properties={"uname": synonym("name"), "ufoo": synonym("uname")},
         )
 
         s = fixture_session()
-        jill = s.query(Foo).get(3)
+        jill = s.get(Foo, 3)
         s.query(Foo).update(
             {Foo.ufoo: "moonbeam"}, synchronize_session="evaluate"
         )
@@ -1490,8 +1526,8 @@ class UpdateDeleteIgnoresLoadersTest(fixtures.MappedTest):
             cls.tables.users,
         )
 
-        mapper(User, users)
-        mapper(
+        cls.mapper_registry.map_imperatively(User, users)
+        cls.mapper_registry.map_imperatively(
             Document,
             documents,
             properties={
@@ -1608,8 +1644,8 @@ class UpdateDeleteFromTest(fixtures.MappedTest):
             cls.tables.users,
         )
 
-        mapper(User, users)
-        mapper(
+        cls.mapper_registry.map_imperatively(User, users)
+        cls.mapper_registry.map_imperatively(
             Document,
             documents,
             properties={"user": relationship(User, backref="documents")},
@@ -1671,11 +1707,37 @@ class UpdateDeleteFromTest(fixtures.MappedTest):
         s = fixture_session()
 
         q = s.query(User).filter(User.id == Document.user_id)
+
         assert_raises_message(
             exc.InvalidRequestError,
             "Could not evaluate current criteria in Python.",
             q.update,
-            {"name": "ed"},
+            {"samename": "ed"},
+        )
+
+    @testing.requires.multi_table_update
+    def test_multi_table_criteria_ok_wo_eval(self):
+        User = self.classes.User
+        Document = self.classes.Document
+
+        s = fixture_session()
+
+        q = s.query(User).filter(User.id == Document.user_id)
+
+        q.update({Document.samename: "ed"}, synchronize_session="fetch")
+        eq_(
+            s.query(User.id, Document.samename, User.samename)
+            .filter(User.id == Document.user_id)
+            .order_by(User.id)
+            .all(),
+            [
+                (1, "ed", None),
+                (1, "ed", None),
+                (2, "ed", None),
+                (2, "ed", None),
+                (3, "ed", None),
+                (3, "ed", None),
+            ],
         )
 
     @testing.requires.update_where_target_in_subquery
@@ -1741,7 +1803,7 @@ class UpdateDeleteFromTest(fixtures.MappedTest):
             ),
         )
 
-    @testing.only_on("mysql", "Multi table update")
+    @testing.requires.multi_table_update
     def test_update_from_multitable_same_names(self):
         Document = self.classes.Document
         User = self.classes.User
@@ -1790,7 +1852,9 @@ class ExpressionUpdateTest(fixtures.MappedTest):
     @classmethod
     def setup_mappers(cls):
         data = cls.tables.data
-        mapper(cls.classes.Data, data, properties={"cnt": data.c.counter})
+        cls.mapper_registry.map_imperatively(
+            cls.classes.Data, data, properties={"cnt": data.c.counter}
+        )
 
     @testing.provide_metadata
     def test_update_attr_names(self):
@@ -2125,7 +2189,7 @@ class LoadFromReturningTest(fixtures.MappedTest):
         User = cls.classes.User
         users = cls.tables.users
 
-        mapper(
+        cls.mapper_registry.map_imperatively(
             User,
             users,
             properties={
